#include "main.hpp"

int Drehregler;  // Ausgabewert des Drehreglers
int speed;  // Das Wort "Geschwindigkeit" steht als Variable für den Ansteuerungswert am ESC.
int last_speed;

time_t run_time = 0;
time_t start_time = 0;
hw_timer_t * timer = NULL;

volatile double integ = 0.0;
volatile double derivative = 0.0;
volatile double esc_output = 0.0;
int ser_esc_output = -1;
double ser_speed = 1000;
volatile double last_error = 0.0;
int esc_value = 0;
volatile double goal_speed = 0;
volatile double current_speed = 0;
volatile double err = 0;
volatile double output = 0.0;
double kp = 0.005;
double ki = 0.0008;
double kd = 0.0001;

//bei 2000 als speed:  values: kp=0.020000 ki=0.005000 kd=0.000100

void ISR_HALL1();
void ISR_HALL2();
void ISR_HALL3();
void ISR_HALL4();
void ISR_HALL5();
void ISR_HALL6();
void ISR_HALL7();
void ISR_HALL8();
void ISR_onTimer();

void data_generate();
void data_store();
void data_check();
void data_init();
void speed_get();
void speed_set();
void PID();
void count_secs(time_t* run_time);

void get_serial_cmd();

#ifdef HAS_MQTT
void json_init();
#endif

void setup() 
{
  delay(1000);
  Serial.begin(115200);
  Serial.println("Init capture portal ...");

  randomSeed(micros());

  display.begin();
  capportal.begin();
  html_content = "<!DOCTYPE html><html><head><meta http-equiv=\"refresh\" content=\"0; url='/_ac'\" /></head><body></body></html>";  

  #ifdef HAS_MQTT
  mqtt.begin(mqttServer, mqttPort, mqttUser, mqttPassword);
  #endif

  Store.begin();
  
  Serial.println("Init ETC ...");
  ESC.attach(ESC_PIN,1000,2000); 

  for(uint8_t i=0;i<=100;i=i+10)
  {
    display.progressBar(i);
    delay(500);
    ESC.write(0);
  }

  data_init();

  // Initialise  timer
  /*
  timer = timerBegin(0, 80, true); // Timer 0, Prescaler 80, zähle im Up-Modus
  timerAttachInterrupt(timer, &ISR_onTimer, true); // Setze den Interrupt-Handler und aktiviere den Interrupt auf steigende Flanke
  timerAlarmWrite(timer, 1000000/8, true); // Setze den Alarm auf 1/4 Sekunde und aktiviere ihn
  timerAlarmEnable(timer); // Aktiviere den Alarm
  */

  pinMode(39, INPUT_PULLDOWN);
  pinMode(HALL1_PIN, INPUT);
  pinMode(HALL2_PIN, INPUT);
  pinMode(HALL3_PIN, INPUT);
  pinMode(HALL4_PIN, INPUT);
  pinMode(HALL5_PIN, INPUT);
  pinMode(HALL6_PIN, INPUT);
  pinMode(HALL7_PIN, INPUT);
  pinMode(HALL8_PIN, INPUT);

  //Serial.println("Init light sensor interrupt ...");
  //attachInterrupt(LS1_PIN, ISR_LS1, RISING);

  Serial.println("Init hall sensor interrupts ...");
  attachInterrupt(digitalPinToInterrupt(HALL1_PIN), ISR_HALL1, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL2_PIN), ISR_HALL2, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL3_PIN), ISR_HALL3, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL4_PIN), ISR_HALL4, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL5_PIN), ISR_HALL5, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL6_PIN), ISR_HALL6, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL7_PIN), ISR_HALL7, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL8_PIN), ISR_HALL8, FALLING);  

  Serial.println(" Auf gehts ... ");
 
}

void loop() 
{

  capportal.handle();
  get_serial_cmd();
  data_generate();
  count_secs(&run_time);
  speed_get();
  speed_set();
  data_check();
  data_store();

  #ifdef HAS_MQTT
  mqtt.send_and_loop("infinity/data", json.create_json_string(AllData, ALL_DATA_COUNT, run_time));
  #endif
  
  display.show_values(speed, MIN_SPEED, MAX_SPEED, HallData, HALL_SENSORS_COUNT, run_time, capportal.localIP());
 
}

void count_secs(time_t* run_time)
{
  static time_t last_update_ms = 0;
  if(last_update_ms == 0)
  {
    last_update_ms = millis();
  }
  if(millis() - last_update_ms >= 1000)
  {
    last_update_ms = millis();
    *run_time = *run_time + 1;
  }
}

double speed_map(double x, double in_min, double in_max, double out_min, double out_max) {
    const double dividend = out_max - out_min;
    const double divisor = in_max - in_min;
    const double delta = x - in_min;
    if(divisor == 0){
        log_e("Invalid map input range, min == max");
        return -1; //AVR returns -1, SAM returns 0
    }
    return (delta * dividend + (divisor / 2)) / divisor + out_min;
}

void PID()
{
  current_speed = HallData[0].value;

  //current_speed = map(HallData[0].value, MIN_SPEED, MAX_SPEED, MIN_ESC_SPEED, MAX_ESC_SPEED); 

  double int_err = goal_speed - current_speed;
  double int_integ = integ + int_err;
  double int_derivative = (int_err - last_error);
  double int_output = kp * int_err + ki * int_integ + kd * int_derivative;
  if((int_output >= MIN_SPEED) && (int_output <= MAX_SPEED))
  {
    err = int_err;
    integ = int_integ;
    derivative = int_derivative;
    output = int_output;
  }

  //Beschränkung der Ausgabe
  esc_output = speed_map(output, MIN_SPEED, MAX_SPEED, MIN_ESC_SPEED, MAX_ESC_SPEED);

  //Serial.printf("PID ... (goal: %f, current: %06.3f, err: %06.3f, integ: %06.3f, derive: %06.3f) output=%06.3f \n", goal_speed, current_speed, err, integ, derivative, output);

  //Aktualisierung der letzten Fehler- und Zeitwerte
  last_error = err;
}



void speed_set()
{
  static unsigned long last_esc_write_ms = 0;
  if(0 == last_esc_write_ms || millis() - last_esc_write_ms > UPDATE_ESC_EVERY_MS)
  {
    PID();
    Serial.printf(">goal:%06.3f\n>current:%06.3f\n>err:%06.3f\n>integ:%06.3f\n>derive:%06.3f\n>output:%06.3f\n>esc_output:%06.3f\n", goal_speed, current_speed, err, integ, derivative, output, esc_output);
    if(-1 == ser_esc_output )
    {
      ESC.write(esc_output);
    }
    else
    {
      ESC.write(ser_esc_output);
    }
    last_esc_write_ms = millis();
  }
}

void speed_get()
{
  static uint8_t _count_meas = 0;
  static int _Drehregler = 0;
  static unsigned long _last_read_ms = 0;
  if(ser_speed == -1)
  {
    if( millis() - _last_read_ms > 25)
    {
      _last_read_ms = millis();
      int maxDrehregler = 0;
      int minDrehregler = 0;
      int sumDrehregler = 0;
      for(uint8_t i = 0; i<6;i++)
      {
        Drehregler = analogRead(POTI_PIN);   // Dieser Befehl liest den Wert des Potentiometers am analogen Pin A0 aus und speichert ihn unter der Variablen "Drehregler". Der Wert liegt zwischen 0 und 1023.
        if(i==0)
        {
          maxDrehregler = Drehregler;
          minDrehregler = Drehregler;
        }
        else
        {
          if(Drehregler < minDrehregler)
          {
            minDrehregler = Drehregler;
          }
          if(Drehregler > maxDrehregler)
          {
            maxDrehregler = Drehregler;
          }
        }
        sumDrehregler = sumDrehregler + Drehregler;
      }
      sumDrehregler = sumDrehregler - maxDrehregler - minDrehregler;
      sumDrehregler = sumDrehregler/4;
      #ifdef DUMMY_DATA
      speed = 1000;
      #else
      speed = map(sumDrehregler, 0, MAX_POTI_VALUE, MIN_SPEED, MAX_SPEED);   // Der "MAP-" Befehl wandelt den Messwert aus der Variablen "Drehregler" um, damit er am ESC verarbeitet werden kann. Der Zahlenbereich 0 bis 1023 wird dabei in einen Zahlenwert zwischen 0 und 180 umgewandelt.
      #endif
    }
  }
  else{
    speed = ser_speed;
  }
  goal_speed = 1.0 * speed;

}

void data_init()
{
  Serial.println("Init internal data ...");
  #ifdef MACHINE_RESTART
    run_time = 0;
    start_time = capportal.epochTime();
    Store.set_int(EEPROM_NAME_RUN_TIME , run_time );
    Store.set_int(EEPROM_NAME_START_EPOCHTIME, start_time );
    Serial.printf("Machine restarted. Set runtime to 0. Stored this value and the epoch start time %lu to EEPROM.\n", start_time);
  #else
    run_time = Store.get_int(EEPROM_NAME_RUN_TIME );
    start_time = Store.get_int(EEPROM_NAME_START_EPOCHTIME );
    Serial.printf("Machine runtime loaded from EPPROM: %lu seconds. Machine was started %s\n", run_time, capportal.localTime(start_time));
  #endif
  last_speed = 0;
  speed = 0;
  unsigned long current_millis = millis();
  for (uint8_t i = 0; i < ALL_DATA_COUNT; i ++)
  {
    AllData[i].name = Data_names[i];
    AllData[i].unit = Data_units[i];
    AllData[i].value = 0;
    AllData[i].unit_factor = Data_units_factor[i];
    AllData[i].period_ticks = 0;
    #ifdef MACHINE_RESTART
    AllData[i].ticks = 0;
    AllData[i].timestamp = 0;
    #else
    AllData[i].ticks = Store.get_int(EEPROM_NAME_TICK+String(i));
    AllData[i].timestamp = Store.get_int(EEPROM_NAME_TIMESTAMP+String(i));
    if(i != 0)
    {
      AllData[i].value = (AllData[i].ticks * AllData[i].unit_factor);
    }
    #endif
    AllData[i].act_update_ms = 0;
    AllData[i].prev_update_ms = 0;    
  }
}


void data_check()
{
  /*
  Serial.printf("H1(pin %d): %d, H2(pin %d): %d, H3(pin %d): %d, H4(pin %d): %d, H5(pin %d) : %d, H6(pin %d) : %d, H7(pin %d) : %d, H8(pin %d) : %d.\n",
                HALL1_PIN, digitalRead(HALL1_PIN),
                HALL2_PIN, digitalRead(HALL2_PIN),
                HALL3_PIN, digitalRead(HALL3_PIN),
                HALL4_PIN, digitalRead(HALL4_PIN),
                HALL5_PIN, digitalRead(HALL5_PIN),
                HALL6_PIN, digitalRead(HALL6_PIN),
                HALL7_PIN, digitalRead(HALL7_PIN),
                HALL8_PIN, digitalRead(HALL8_PIN)                
                );
  
  */

  for(uint8_t nr = 0; nr < HALL_SENSORS_COUNT; nr++)
  {
    if( 0 == nr)
    {
      if( (HallData[nr].act_update_ms != 0)
       && (HallData[nr].prev_update_ms != 0)
       && (HallData[nr].prev_update_ms != HallData[nr].act_update_ms)
        )
      {
        HallData[nr].value = HallData[nr].unit_factor / ((HallData[nr].act_update_ms - HallData[nr].prev_update_ms)/ HALL_NR_TURNS) ;
      }
    }
    else{
      HallData[nr].value = (HallData[nr].ticks * HallData[nr].unit_factor);
    }
  }
}

void data_store()
{
  static unsigned long _last_stored_ms = 0;
  if(_last_stored_ms == 0)
  {
    _last_stored_ms = millis();
  }
  if(millis() - _last_stored_ms >= EEPROM_STORE_RUNTIME_ALL_MS)
  {
    _last_stored_ms = millis();
    Store.set_int(EEPROM_NAME_RUN_TIME, run_time);
    for(uint8_t nr = 0; nr < HALL_SENSORS_COUNT; nr++)
    {
      Store.set_int(EEPROM_NAME_TICK+String(nr), HallData[nr].ticks);
      Store.set_int(EEPROM_NAME_TIMESTAMP+String(nr), HallData[nr].timestamp);
    }
  }
}


void get_serial_cmd()
{
  if(Serial.available()){
        String command = Serial.readStringUntil('\n');
        Serial.printf("Received command '%s'", command.c_str());
        if(command.startsWith("kp=")){
            kp = command.substring(command.indexOf("=")+1).toDouble();
            Serial.printf("Set kp to %f\n", kp);
        }
        else if(command.startsWith("ki=")){
            ki = command.substring(command.indexOf("=")+1).toDouble();
            Serial.printf("Set ki to %f\n", ki);
        }
        else if(command.startsWith("speed=")){
            ser_speed = command.substring(command.indexOf("=")+1).toDouble();
            Serial.printf("Set speed to %f\n", ser_speed);
        }
        else if(command.startsWith("kd=")){
            kd = command.substring(command.indexOf("=")+1).toDouble();
            Serial.printf("Set kd to %f\n", kd);
        }
        else if(command.startsWith("esc=")){
            ser_esc_output = command.substring(command.indexOf("=")+1).toDouble();
            Serial.printf("Set esc_output to %f\n", ser_esc_output);
        }

        else if(command.equals("params")){
            Serial.printf("Actual values: kp=%f  ki=%f  kd=%f  \n", kp, ki, kd);
        }
        else if(command.equals("control")){
            ser_esc_output = -1;
            Serial.printf("Set to drive PID control mode.\n");
        }
        else if(command.equals("max")){
            ser_esc_output = MAX_ESC_SPEED;
            Serial.printf("Set to maximal speed. %f\n", ser_esc_output);
        }        
        else if(command.equals("min")){
            ser_esc_output = MIN_ESC_SPEED;
            Serial.printf("Set to minimal speed %f .\n", ser_esc_output );
        }
        else if(command.equals("start")){
            ser_esc_output = -1;
            Serial.printf("Start drive PID control mode.\n");
        }
        else if(command.equals("stop")){
            ser_esc_output = 0;
            Serial.printf("Stop drive.\n");
        }
        else if(command.equals("help")){
            Serial.printf("Commands: kp,ki,kd=x.xxxx | params | control | max | min | start | stop | speed=x | esc=x help \n");
        }
        else{
            Serial.printf("Invalid command.");
        }
  }
}

void IRAM_ATTR ISR_HALL1()
{
  const uint8_t hallnr = 0;
  HallData[hallnr].ticks++;
  HallData[hallnr].period_ticks++;
  HallData[hallnr].timestamp = run_time;
  // each 4 ticks is one turn
  // so we calculate speed each 4 ticks... just in case we check for bigger ... 
  if(HallData[hallnr].period_ticks >= HALL_NR_TURNS*HALL_TICKS_PER_TURN){
    HallData[hallnr].period_ticks = 0;
    HallData[hallnr].prev_update_ms = HallData[hallnr].act_update_ms;
    HallData[hallnr].act_update_ms = millis();
  }
}

void IRAM_ATTR ISR_HALL2()
{
  const uint8_t hallnr = 1;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

void IRAM_ATTR ISR_HALL3()
{
  const uint8_t hallnr = 2;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

void IRAM_ATTR ISR_HALL4()
{
  const uint8_t hallnr = 3;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

void IRAM_ATTR ISR_HALL5()
{
  const uint8_t hallnr = 4;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

void IRAM_ATTR ISR_HALL6()
{
  const uint8_t hallnr = 5;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

void IRAM_ATTR ISR_HALL7()
{
  const uint8_t hallnr = 6;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

void IRAM_ATTR ISR_HALL8()
{
  const uint8_t hallnr = 7;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

//this code is just for generating dummy data - in normal use not needed
void data_generate()
{
  #ifdef DUMMY_DATA

  static unsigned long _last_updated_ms = 0; 
  speed = 1000 + random(-250,150);
  unsigned long _wait_ms = map( speed, MIN_SPEED, MAX_SPEED, 60000/MIN_SPEED/4, 60000/MAX_SPEED/4);
  //Serial.printf("Speed is : %lu and _wait_ms is %lu\n", speed, _wait_ms);

  if(_last_updated_ms == 0)
  {
    _last_updated_ms = millis();
  }

  if(millis() - _last_updated_ms >= _wait_ms )
  {
    _last_updated_ms = millis();

    for(uint8_t i = 0; i < HALL_SENSORS_COUNT; i++)
    {
      if(i == 0)
      {
        HallData[i].ticks++;
        HallData[i].period_ticks++;
        HallData[i].last_update_ms = millis();
        HallData[i].timestamp = run_time;
      }
      else
      {
        //Serial.printf("i=%d HallData[i-1].ticks=%lu %4 = %d\n" , i, HallData[i-1].ticks , HallData[i-1].ticks % 4);
        if (HallData[i].last_update_ms != HallData[i-1].last_update_ms && HallData[i-1].ticks % 4 == 0)
        {
          HallData[i].ticks++;
          HallData[i].period_ticks++;
          HallData[i].last_update_ms = HallData[i-1].last_update_ms;
          HallData[i].timestamp = run_time;
        }
      }
    }
  }
  #endif
}
